#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <gst/gst.h>
#include <gnet.h>
#include <glib.h>
#include <time.h>

#define AUDIO_DEPAY "rtppcmadepay"
#define AUDIO_PAY "rtppcmapay"
#define AUDIO_DEC   "alawdec"
#define AUDIO_ENC   "alawenc"
#define AUDIO_SINK  "alsasink"

#define MULTI_HOST "158.110.29.255"
#define OUT_PORT 55000
#define SERVICE_PORT 60000

enum status_enum 
{
   OFF,
   AUTHENTICATED,
   ACTIVE,
};


typedef struct member_struct
{
    enum status_enum status;
    char nickname[32];
    GInetAddr* InetAddr;
    uint ssrc;
    GstElement* mybin;
    GstElement* amplifier;
} member_var;

/* callbacks  */
static gboolean my_bus_callback (GstBus *bus, GstMessage *message, gpointer data);
static gboolean receive_cb (GIOChannel* channel,GIOCondition cond,gpointer data);
static void pad_added_cb (GstElement * rtpbin, GstPad * new_pad, gpointer data);
static void pad_removed_cb (GstElement * rtpbin, guint session, guint ssrc, gpointer data);
/* functions  */
uint new_ssrc();
void remove_member_elements(int i);
int members_active();
void set_amplification();


/* global var   */
GstElement *pipeline;
member_var member[8];
GUdpSocket* server;
uint rtpport=50000,rtcpport=50001;

/* main function   */
int main (int argc, char *argv[])
{

    /* loop and bus */
    GMainLoop *loop;
    GstBus *bus;
    /* common pipeline elements */
    GstElement *rtpbin, *rtpsrc;
    
    #ifdef RTCP
      GstElement *rtcpsrc;
    #endif
    
    GstElement *adder, *queue, *enc, *rtppay, *rtpsink;
    /* network */
    GIOChannel *channel;
    /* useful items */
    GstCaps *caps;
    GstPad *srcpad, *sinkpad;
    GstPadLinkReturn lres;
    gboolean res;
    int i;


    /* always init first */
    gst_init (&argc, &argv);
    gnet_init ();
    loop = g_main_loop_new (NULL, FALSE);
    for (i=0;i<=7;i++)
        member[i].status=OFF;

    /* Create the socket  */
    server = gnet_udp_socket_new_with_port (SERVICE_PORT);
    g_assert (server);
    channel=gnet_udp_socket_get_io_channel(server);
    g_assert(channel);

    /* the pipeline to hold everything */
    pipeline = gst_pipeline_new (NULL);
    g_assert (pipeline);

     /* udp source streams */
    rtpsrc = gst_element_factory_make ("udpsrc", "rtpsrc");
    g_assert (rtpsrc);
    
    g_object_set (rtpsrc, "port", rtpport, NULL);
/*    gst_element_set_state (rtpsrc, GST_STATE_PAUSED);*/
/*    g_object_get (rtpsrc,"port", &rtpport,NULL);*/
    
    #ifdef RTCP
    rtcpsrc = gst_element_factory_make ("udpsrc", "rtcpsrc");
    g_assert (rtcpsrc);

    g_object_set (rtcpsrc, "port", rtcpport, NULL);
/*    gst_element_set_state (rtcpsrc, GST_STATE_PAUSED);*/
/*    g_object_get (rtcpsrc,"port", &rtcpport,NULL);*/
    #endif

    /* we need to set caps on the udpsrc for the RTP data */
    caps = gst_caps_new_simple ("application/x-rtp",
				"media",	G_TYPE_STRING, 	"audio",
				"clock-rate",	G_TYPE_INT, 	8000,
				"channels", 	G_TYPE_INT, 	1,
				"payload", 	G_TYPE_INT, 	8,
				"encoding-name",G_TYPE_STRING,  "PCMA",
				NULL);
    g_object_set (rtpsrc, "caps", caps, NULL);
    gst_caps_unref (caps);
    /*****************************************************/


    /* for mixing */
    adder = gst_element_factory_make ("adder", NULL);
    g_assert (adder);

    /* the rtpbin element */
    rtpbin = gst_element_factory_make ("gstrtpbin", NULL);
    g_assert (rtpbin);
    g_object_set (G_OBJECT(rtpbin), "autoremove", TRUE, NULL);

    /* queue to seperate thread   */
    queue = gst_element_factory_make ("queue", NULL);
    g_assert (queue);

    /* encoder e rtppay */
    enc = gst_element_factory_make (AUDIO_ENC, NULL);
    g_assert (enc);
    rtppay = gst_element_factory_make (AUDIO_PAY, NULL);
    g_assert (rtppay);

    /* audiosink ******************************************/
    rtpsink = gst_element_factory_make ("udpsink", NULL);
    g_assert (rtpsink);
    g_object_set (rtpsink, "sync", FALSE, NULL);
    g_object_set (rtpsink, "async", FALSE, NULL);
    g_object_set (rtpsink, "host", MULTI_HOST,"port",OUT_PORT, NULL);

    /* add elements to the pipeline   */
    gst_bin_add_many (GST_BIN (pipeline), rtpsrc, rtpbin, adder, queue, enc, rtppay, rtpsink, NULL);
    #ifdef RTCP
    gst_bin_add (GST_BIN(pipeline),rtcpsrc);
    #endif

    /* now link what we can link */
    srcpad = gst_element_get_static_pad (rtpsrc, "src");
    sinkpad = gst_element_get_request_pad (rtpbin, "recv_rtp_sink_0");
    lres = gst_pad_link (srcpad, sinkpad);
    g_assert (lres == GST_PAD_LINK_OK);
    gst_object_unref (srcpad);
    gst_object_unref (sinkpad);

    #ifdef RTCP
    /* we also want to receive RTCP, request an RTCP sinkpad for session 0 and */
    /* link it to the srcpad of the udpsrc for RTCP */
    srcpad = gst_element_get_static_pad (rtcpsrc, "src");
	sinkpad = gst_element_get_request_pad (rtpbin, "recv_rtcp_sink_0");
    lres = gst_pad_link (srcpad, sinkpad);
    g_assert (lres == GST_PAD_LINK_OK);
    gst_object_unref (srcpad);
    gst_object_unref (sinkpad);
    #endif

    res=gst_element_link_many(adder,queue,enc,rtppay,rtpsink,NULL);
    g_assert(res==TRUE);

    /* set the bus */
    bus = gst_pipeline_get_bus (GST_PIPELINE (pipeline));
    gst_bus_add_watch (bus, my_bus_callback, loop);
    gst_object_unref (bus);


    /* the RTP pad that we have to connect to the depayloader will be created
    * dynamically so we connect to the pad-added signal, pass the depayloader as
    * user_data so that we can link to it. */
    g_signal_connect (rtpbin, "pad-added", G_CALLBACK (pad_added_cb), adder);
    /* g_signal_connect (new_pad, "unlinked", G_CALLBACK (pad_removed_cb), NULL);*/
    g_signal_connect (rtpbin, "on-sender-timeout", G_CALLBACK (pad_removed_cb), NULL);

    /* add watch to the channel  */
    g_io_add_watch (channel,(G_IO_IN | G_IO_ERR),receive_cb,NULL);

    /* set the pipeline to playing */
    gst_element_set_state (pipeline, GST_STATE_PLAYING);
    g_print("listening for udp messages on port: %d\n",SERVICE_PORT);
    g_print("listening for rtp streams on port: %d\n",rtpport);
    #ifdef RTCP
    g_print("listening for rtcp streams on port: %d\n",rtcpport);
    #endif
	g_print("rtp out streams: broadcast on port %d\n",OUT_PORT);
    g_print("server is running...\n");
    g_main_loop_run (loop);

    /* when the  loop stop   */
    g_print ("stopping server pipeline\n");
    gst_element_set_state (pipeline, GST_STATE_NULL);

    gst_object_unref (pipeline);

    return EXIT_SUCCESS;
}


/******************************************************************************/

static gboolean my_bus_callback (GstBus *bus, GstMessage *message, gpointer data)
{
    GMainLoop *loop = data;

    if (GST_MESSAGE_TYPE (message)==GST_MESSAGE_ERROR)
    {
        GError *err;
        gchar  *debug;

        gst_message_parse_error (message, &err, &debug);
        g_print ("Error: %s\n", err->message);
        g_error_free (err);
        g_free (debug);

        g_main_loop_quit (loop);
        return TRUE;
    }
    else if (GST_MESSAGE_TYPE (message)==GST_MESSAGE_EOS)
    {
        /* end-of-stream */
        g_main_loop_quit (loop);
        return TRUE;
    }

    return TRUE;
}


static void pad_added_cb (GstElement * rtpbin, GstPad * new_pad, gpointer data)
{
    GstElement *adder=data;
    GstElement *mybin,*fake;
    GstElement *rtpdepay, *audiodec, *audioconv, *amplifier;
    GstPad *sinkpad,*srcpad;
    GstPadLinkReturn lres;
    gboolean res;
    uint ssrc;
    int i=0,session;

    g_print ("new payload on pad: %s\n", GST_PAD_NAME (new_pad));

    /* there is a new stream get the ssrc */
    if (sscanf(GST_PAD_NAME(new_pad),"recv_rtp_src_%d_%d_",&session,&ssrc)==-1)
    {
        g_print("cannot read the new ssrc\n");
        return;
    }

    /* check if the ssrc is authenticated */
    i=0;
    while (((member[i].ssrc!=ssrc)||(member[i].status!=AUTHENTICATED))&&(i<=7))
        i++;
    if (i>7)
    {
        g_print("cannot find ssrc or ssrc is not authenticated, connecting a fakesink to the new pad... ");
        fake=gst_element_factory_make("fakesink",NULL);
        g_assert(fake);
        gst_bin_add(GST_BIN(pipeline),fake);
        sinkpad=gst_element_get_pad(fake,"sink");
        g_assert(sinkpad);
        lres = gst_pad_link (new_pad, sinkpad);
        g_assert (lres == GST_PAD_LINK_OK);
        gst_object_unref(sinkpad);
        gst_element_set_state (fake, GST_STATE_PAUSED);
        g_print(" done\n");
        return;
    }

    g_print("%s (authenticated) join the session, connecting elements... ",member[i].nickname);

    /* the depayloading and decoding */
    rtpdepay= gst_element_factory_make (AUDIO_DEPAY, NULL);
    g_assert (rtpdepay);
    audiodec = gst_element_factory_make (AUDIO_DEC, NULL);
    g_assert (audiodec);
    /* format conversion and amplification */
    audioconv = gst_element_factory_make ("audioconvert", NULL);
    g_assert (audioconv);
    amplifier = gst_element_factory_make ("audioamplify", NULL);
    g_assert(amplifier);
    g_object_set(amplifier,"amplification",1.0,NULL);

    /* put all elements in a new bin  */
    mybin=gst_bin_new(NULL);
    gst_bin_add_many (GST_BIN (mybin), rtpdepay, audiodec, audioconv, amplifier, NULL);

    /* link the elements in mybin   */
    res = gst_element_link_many (rtpdepay, audiodec, audioconv, amplifier, NULL);
    g_assert (res == TRUE);

    /* now add mybin to the pipeline   */
    gst_bin_add(GST_BIN (pipeline), mybin);


    /* create sink and source ghostpads on mybin    */
    sinkpad = gst_element_get_static_pad (rtpdepay, "sink");
    g_assert(sinkpad);
    res=gst_element_add_pad (mybin, gst_ghost_pad_new ("sink", sinkpad));
    g_assert(res==TRUE);
    gst_object_unref(sinkpad);

    srcpad = gst_element_get_static_pad (amplifier, "src");
    g_assert(srcpad);
    res=gst_element_add_pad (mybin, gst_ghost_pad_new ("src", srcpad));
    g_assert(res==TRUE);
    gst_object_unref(srcpad);

    /* connect the new_pad to the rtpdepay (mybin)    */
    sinkpad = gst_element_get_static_pad (mybin, "sink");
    g_assert (sinkpad);
    lres = gst_pad_link (new_pad, sinkpad);
    g_assert (lres == GST_PAD_LINK_OK);
    gst_object_unref (sinkpad);

    /* connect the audioconv (mybin) to the adder    */
    res = gst_element_link (mybin, adder);
    g_assert (res == TRUE);

    /* last things  *****/
    gst_element_set_state (mybin, GST_STATE_PAUSED);

    /* update the data of the new member */
    member[i].mybin=mybin;
    gst_object_ref (member[i].mybin);
    member[i].amplifier=amplifier;
    //gst_object_ref (member[i].amplifier);
    member[i].status=ACTIVE;
    set_amplification();
    g_print(" done\n");
    g_print("member id: %d, ssrc: %d number of active members: %d\n",i,member[i].ssrc,members_active());

    return;
}

static void pad_removed_cb (GstElement* rtpbin, guint session, guint ssrc, gpointer data)
{
    int i;
    GstElement *adder,*fake;
    GstPad *sinkpad,*srcpad;
    gchar name[32];

    sprintf(name,"recv_rtp_src_%d_%u_8",session,ssrc);
    srcpad=gst_element_get_static_pad(rtpbin,name);

    /* get the member of the callback */
    i=0;
    while (((member[i].ssrc!=ssrc)||(member[i].status!=ACTIVE))&&(i<=7))
        i++;

    if (i>=8)  /* ssrc not found so unlink the fakesrc */
    {
        g_print("a fake-member or a member that announced the logout is gone in timeout,\nremoving fakesink... ");
        sinkpad=gst_pad_get_peer(srcpad);
        gst_pad_unlink(srcpad,sinkpad);

        fake=gst_pad_get_parent_element(sinkpad);
        gst_element_set_state (fake, GST_STATE_NULL);
        gst_bin_remove(GST_BIN(pipeline), fake);
        gst_object_unref(sinkpad);
        gst_object_unref(srcpad);
        gst_object_unref(fake);
        
        g_print(" done\n");
        return;
    }
    else  /* ohps! */
    {
        g_print("member: %s has left the session without announce, restoring stream... ",member[i].nickname);
        /* check if the pad is unlinked otherwise unlink  */
        if (gst_pad_is_linked(srcpad))
        {
            sinkpad=gst_pad_get_peer(srcpad);
            gst_pad_unlink(srcpad,sinkpad);
            gst_object_unref(sinkpad);
            gst_object_unref(srcpad);
        }

        /* unlink mybin and adder and remove mybin */
        srcpad=gst_element_get_pad(member[i].mybin,"src");
        sinkpad=gst_pad_get_peer(srcpad);
        gst_pad_unlink(srcpad,sinkpad);

        gst_element_set_state (member[i].mybin, GST_STATE_NULL);
        gst_bin_remove(GST_BIN(pipeline), member[i].mybin);
        gst_object_unref(srcpad);
        //gst_object_unref(member[i].mybin);  no need to unref

        adder=gst_pad_get_parent_element(sinkpad);
        gst_element_release_request_pad (adder,sinkpad);
        gst_object_unref(sinkpad);
        gst_object_unref(adder);
        
        /*  Set to READY State the Pipeline  */
/*        gst_element_set_state(pipeline, GST_STATE_READY);*/
        /*  Set to PAUSED State the Pipeline */
/*        gst_element_set_state (pipeline, GST_STATE_PAUSED);*/
        /*  Set to NULL State the Pipeline  */
/*        gst_element_set_state(pipeline, GST_STATE_NULL);*/
        /*  Set to PLAY State the Pipeline  */

/*        gst_element_set_state (pipeline, GST_STATE_PLAYING);*/
        
        member[i].status=OFF;
        member[i].ssrc=0;
        gnet_inetaddr_unref (member[i].InetAddr);
        set_amplification();
        g_print(" done\nmembers remaining %d\n",members_active());
        member[i].nickname[0]='\0';

       
        return;
    }      
}

static gboolean receive_cb (GIOChannel* channel,GIOCondition cond,gpointer data)
{
    char buffer[128]={0};
    char *ip,temp[64];
    int i,port;
    //uint n;
    GInetAddr* temp_addr;
    uint ssrc;
    
    temp_addr = gnet_inetaddr_new ("255.255.255.255",65535);

    if (cond==G_IO_ERR)
    {
        g_print("error occurred on the udp socket\n");
        exit(EXIT_FAILURE);
    }
    
    gnet_udp_socket_receive (server,buffer,128,&temp_addr);
   

    if (strncmp(buffer,"ENQ:AUTCT",9)==0)
    {
        if(sscanf(&(buffer[9]),"%s",temp)!=1)
		    {
			    g_print("error cannot read nickname on an authentication request\n");
			    gnet_udp_socket_send(server,"ERR:UNK_NKN",12,temp_addr);
			    return TRUE;
		    }

        i=0;
        while ((member[i].status!=OFF)&&(i<=7))
            i++;

        if (i>7)
        {
            gnet_udp_socket_send(server,"ERR:SVR_FULL",13,temp_addr); //tu reinvii a quest' indirizzo, ma non è detto che la porta sia libera, se è quella sull quale è arrivato a te
            g_print("a connection request was refused: server full\n");
            return TRUE;
        }

		ip=gnet_inetaddr_get_canonical_name(temp_addr);
		port=gnet_inetaddr_get_port(temp_addr);
		
        member[i].InetAddr=gnet_inetaddr_clone(temp_addr);
        gnet_inetaddr_unref (temp_addr);
        member[i].ssrc=new_ssrc();
        member[i].nickname[0]='\0';
        strcpy(member[i].nickname,temp);
        //g_print("%s\n",member[i].nickname);
        member[i].status=AUTHENTICATED;

        g_print("authentication request by %s @ %s:%d\nnew member ssrc: %u\n",temp,ip,port,member[i].ssrc);
        
        // g_free ip
        
        sprintf(temp,"OK:MBR_AUTCT %u",member[i].ssrc);
        gnet_udp_socket_send(server,temp,strlen(temp),member[i].InetAddr);
        return TRUE;
    }
    else if (strncmp(buffer,"ENQ:DSCONN",10)==0)
    {
        if(sscanf(&(buffer[11]),"%d",&ssrc)==-1)
        {
            g_print("cannot read ssrc on a disconnection request\n");
			      gnet_udp_socket_send(server,"ERR:UNK_SSRC",13,temp_addr);
            return TRUE;
        }

        i=0;
        while (!((member[i].status!=OFF)&&(member[i].ssrc==ssrc))&&(i<=7))
            i++;

        if (i>7)
        {
            g_print("error cannot find ssrc on a disconnection request\n");
			
            return TRUE;
        }

        if (member[i].status==ACTIVE)
        {
            g_print("%s send a logout request, removing elements... ",member[i].nickname);
            remove_member_elements(i);
            g_print(" done\n");
        }

        member[i].status=OFF;
        member[i].ssrc=0;
        gnet_udp_socket_send(server,"OK:MBR_DSCONN",13,member[i].InetAddr);
        gnet_inetaddr_unref (member[i].InetAddr);
        set_amplification();
        g_print("%d active members remaining\n",members_active());
        member[i].nickname[0]='\0';
        return TRUE;
    }
    else if (strncmp(buffer,"ENQ:INF:ADDR",12)==0)
    {
        #ifdef RTCP
        sprintf(buffer,"INF:ADDR rtpin=%u rtpout=%s:%u",rtpport,MULTI_HOST,OUT_PORT);
        #else 
        sprintf(buffer,"INF:ADDR rtpin=%u rtcpin=%u rtpout=%s:%u",rtpport,rtcpport,MULTI_HOST,OUT_PORT);
        #endif

        gnet_udp_socket_send(server,buffer,strlen(buffer),temp_addr);
        return TRUE;  
    }
    else 
    {
        g_print("error unknown request received\n");
		gnet_udp_socket_send(server,"ERR:UNK_REQ",12,temp_addr);
		return TRUE;
    }

    return TRUE;
}

uint new_ssrc()
{
    uint ssrc;
    GRand *ran_seed;
    gboolean exit=FALSE;
    int i;

    ran_seed=g_rand_new_with_seed(time(NULL));

    while(!exit)
    {
        ssrc=g_rand_int_range(ran_seed,1,RAND_MAX);
        exit=TRUE;

        for (i=0;i<=7;i++)
            if (member[i].ssrc==ssrc)
                exit=FALSE;
    }

    g_rand_free(ran_seed);
    return ssrc;
}

void remove_member_elements(int i)
{
    GstElement *adder,*fake;
//  GstCaps *caps;
    GstPad *srcpad, *sinkpad;
    GstPadLinkReturn lres;
//  gboolean res;

    sinkpad=gst_element_get_pad(member[i].mybin,"sink");

    /* check if the pad is unlinked otherwise unlink  */
    srcpad=gst_pad_get_peer(sinkpad);
    gst_pad_unlink(srcpad,sinkpad);
    gst_object_unref(sinkpad);
    /* connect the fake element */
    fake=gst_element_factory_make("fakesink",NULL);
    g_assert(fake);
    gst_bin_add(GST_BIN(pipeline),fake);
    sinkpad=gst_element_get_pad(fake,"sink");
    g_assert(sinkpad);
    lres = gst_pad_link (srcpad, sinkpad);
    g_assert (lres == GST_PAD_LINK_OK);
    gst_object_unref(sinkpad);
    gst_object_unref(srcpad);
    gst_element_set_state (fake, GST_STATE_PAUSED);

    /* unlink mybin and adder and remove mybin */
    srcpad=gst_element_get_pad(member[i].mybin,"src");
    sinkpad=gst_pad_get_peer(srcpad);
    gst_pad_unlink(srcpad,sinkpad);

    gst_element_set_state (member[i].mybin, GST_STATE_NULL);
    gst_bin_remove(GST_BIN(pipeline), member[i].mybin);
    gst_object_unref(srcpad);
    //gst_object_unref(member[i].mybin);  no need to unref

    adder=gst_pad_get_parent_element(sinkpad);
    gst_element_release_request_pad (adder,sinkpad);
    gst_object_unref(sinkpad);
    gst_object_unref(adder);

    return;
}

int members_active()
{
    int i,n=0;

    for(i=0;i<=7;i++)
        if (member[i].status==ACTIVE)
            n++;
    return n;
}

void set_amplification()
{
    gfloat amp;
    int i;
    gchar buffer[512]={'\0'};
    gfloat n;
    
    
    n = members_active();
    
    if (n==0) 
        amp=1;
    else      
        amp=1/n;

    sprintf(buffer,"INF:MBR_ACV %d",members_active());
    
    for (i=0;i<=7;i++)
        if (member[i].status==ACTIVE)
        {
            strcat(buffer," ");
            g_object_set(member[i].amplifier,"amplification",amp,NULL);
            strcat(buffer,member[i].nickname);
        }
        
    for (i=0;i<=7;i++)
        if (member[i].status==ACTIVE)
            gnet_udp_socket_send(server,buffer,strlen(buffer),member[i].InetAddr);   

    return;
}
